@@ -240,7 +240,6 @@ def findPaths(
     max_inner_fraction=0.1,
     min_node_n_points=2,
     max_n_points=None,
-    # max_empty_curve_fraction=.2,
     min_compactness=0.5,
     radius=None,
     allow_same_branch=True,
@@ -250,6 +249,7 @@ def findPaths(
     cycle_Lambda=None,
     cycle_Mu=None,
     weights=None,
+    ignore_equivalent=False,
     plot=False,
     verbose=False,
 ):
@@ -279,10 +279,8 @@ def findPaths(
     plot: bool, default=False
         Whether to plot selected candidate paths
     verbose: bool, default=False
-    copy: bool, default=False
-    use_weights: bool, default=False
+    weights: bool, default=False
         Whether to use point weights
-    use_partition: bool or list, default=False
     """
 
     _PG = deepcopy(PG)
@@ -311,9 +309,9 @@ def findPaths(
     if Lambda is None:
         Lambda = _PG["Lambda"]
     if cycle_Mu is None:
-        cycle_Mu = Mu / 10
+        cycle_Mu = Mu 
     if cycle_Lambda is None:
-        cycle_Lambda = Lambda / 10
+        cycle_Lambda = Lambda 
     if radius is None:
         radius = np.mean(edge_lengths) * len(init_nodes_pos) / 10
     if min_path_len is None:
@@ -348,9 +346,9 @@ def findPaths(
         branches = net.get_shortest_paths(leaves[0], leaves[-1])
     else:
         (
-            dict_tree,
+            _,
             dict_branches,
-            dict_branches_single_end,
+            _,
         ) = elpigraph.src.supervised.get_tree(init_edges, leaves[0])
         branches = list(dict_branches.values())
 
@@ -415,9 +413,7 @@ def findPaths(
 
             # ---get nodep, edges, create new graph with added loop
             nodep, edges = pg["NodePositions"], pg["Edges"][0]
-            # _part, _part_dist = elpigraph.src.core.PartitionData(
-            #    X_fit, nodep, 10 ** 6, np.sum(X_fit ** 2, axis=1, keepdims=1)
-            # )
+
             _edges = edges.copy()
             _edges[(edges != 0) & (edges != 1)] += init_edges.max() - 1
             _edges[edges == 0] = c
@@ -436,32 +432,16 @@ def findPaths(
                 cycle_nodes=cycle_nodes,
             )
 
-            (
-                _merged_nodep,
-                _,
-                _,
-                _,
-                _,
-                _,
-                _,
-            ) = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
+            _merged_nodep = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
                 X,
                 _merged_nodep,
                 ElasticMatrix,
                 PointWeights=weights,
                 FixNodesAtPoints=[],
-            )
+            )[0]
 
             ### candidate validity tests ###
             valid = True
-            # --- curve validity test
-            # if (max_empty_curve_fraction is not None) and valid: # if X_fit projected to curve has long gaps
-            #    infer_pseudotime(_adata,source=0)
-            #    sorted_X_proj=_adata.obsm['X_epg_proj'][_adata.obs['epg_pseudotime'].argsort()]
-            #    dist = np.sqrt((np.diff(sorted_X_proj,axis=0)**2).sum(axis=1))
-            #    curve_len = np.sum(_adata.uns['epg']['edge_len'])
-            #    if np.max(dist) > (curve_len * max_empty_curve_fraction):
-            #        valid = False
 
             # --- cycle validity test
             if valid:
@@ -499,7 +479,7 @@ def findPaths(
                         cycle_2d, factor=0.1
                     )
 
-                    # prevent shapely bugs when multi-polygon is returned. Fall back to mahalanobis
+                    # prevent shapely edge case bug when multi-polygon is returned. Fall back to mahalanobis
                     if type(shrunk_cycle_2d) == MultiPolygon:
                         in_shrunk_cycle = np.ones(len(X_inside), dtype=bool)
                     else:
@@ -513,11 +493,6 @@ def findPaths(
                     idx_close = in_shrunk_cycle | (w < 1)
                     w = 1 - w / w.max()
                     w[idx_close] = 1
-
-                    # cycle_nodes_array = np.append(np.array(list(zip(range(len(cycle_2d)-1),
-                    #                                  range(1,len(cycle_2d))))),[[len(cycle_2d)-1,0]],axis=0)
-                    # w, idx_close = get_weights_lineproj(X_inside,cycle_2d,cycle_nodes_array,cycle_centroid[0],threshold=.2)
-
                     inner_fraction = np.sum(w) / np.sum(cycle_points)
 
                 if init_nodes_pos.shape[1] == 2:
@@ -546,8 +521,8 @@ def findPaths(
                     )  # if high fraction of points inside
                     or (not np.isfinite(inner_fraction))  # prevent no points error
                     or (np.sum(idx_close) > max_n_points)  # if too many points inside
-                    or pp_compactness(cycle_2d) < min_compactness
-                ):  # if loop is very narrow
+                    or pp_compactness(cycle_2d) < min_compactness # if loop is very narrow
+                ):  
                     valid = False
 
             # ---> if cycle is invalid, continue
@@ -571,11 +546,6 @@ def findPaths(
                     X, _merged_nodep, _merged_edges, _merged_part
                 )
                 MSE = proj["MSEP"]
-                # dist2proj = np.sum(np.square(X - X_proj), axis=1)
-                # ElasticMatrix = elpigraph.src.core.Encode2ElasticMatrix(_merged_edges, Lambdas=Lambda, Mus=Mu)
-                # ElasticEnergy, MSE, EP, RP = elpigraph.src.core.ComputePenalizedPrimitiveGraphElasticEnergy(_merged_nodep,
-                #                                                                                            ElasticMatrix,
-                #                                                                                            dist2proj,alpha=0.01,beta=0)
                 inner_fractions.append(inner_fraction)
                 energies.append(MSE)
                 merged_edges.append(_merged_edges)
@@ -682,40 +652,46 @@ def findPaths(
 
     # ignore equivalent loops (with more than 2/3 shared points and nodes)
     valid = np.ones(len(new_part))
-    for i in range(len(new_part) - 1):
-        for j in range(i + 1, len(new_part)):
-            if (
-                len(np.intersect1d(new_part[i], new_part[j]))
-                / max(len(new_part[i]), len(new_part[j]))
-            ) > (2 / 3):
-                if np.argmin([new_inner_fraction[i], new_inner_fraction[j]]) == 0:
-                    valid[i] = 0
-                else:
-                    valid[j] = 0
 
-    new_edges = [e for i, e in enumerate(new_edges) if valid[i]]
-    new_nodep = [e for i, e in enumerate(new_nodep) if valid[i]]
-    new_leaves = [e for i, e in enumerate(new_leaves) if valid[i]]
-    new_part = [e for i, e in enumerate(new_part) if valid[i]]
-    new_energy = [e for i, e in enumerate(new_energy) if valid[i]]
-    new_inner_fraction = [e for i, e in enumerate(new_inner_fraction) if valid[i]]
+    if ignore_equivalent:
+        for i in range(len(new_part) - 1):
+            for j in range(i + 1, len(new_part)):
+                if (
+                    len(np.intersect1d(new_part[i], new_part[j]))
+                    / max(len(new_part[i]), len(new_part[j]))
+                ) > (2 / 3):
+
+                    #all_leaves_i = np.isin(new_leaves[i],leaves).all()
+                    #all_leaves_j = np.isin(new_leaves[j],leaves).all()
+                    ## prioritize connecting leaves
+                    #if all_leaves_j and not(all_leaves_i):
+                    #    valid[i] = 0
+                    #elif all_leaves_i and not(all_leaves_j):
+                    #    valid[j] = 0
+                    # else take lowest energy
+                    if np.argmin([new_energy[i], new_energy[j]]) == 0:
+                        valid[j] = 0
+                    else:
+                        valid[i] = 0
+
+        new_edges = [e for i, e in enumerate(new_edges) if valid[i]]
+        new_nodep = [e for i, e in enumerate(new_nodep) if valid[i]]
+        new_leaves = [e for i, e in enumerate(new_leaves) if valid[i]]
+        new_part = [e for i, e in enumerate(new_part) if valid[i]]
+        new_energy = [e for i, e in enumerate(new_energy) if valid[i]]
+        new_inner_fraction = [e for i, e in enumerate(new_inner_fraction) if valid[i]]
 
     ### form graph with all valid loops found ###
     if (new_edges == []) or (sum(valid) == 0):
         print("Found no valid path to add")
         return None
 
+    merged_edges=init_edges.copy()
     for i, loop_edges in enumerate(new_edges):
-        if i == 0:
-            loop_edges[(loop_edges != 0) & (loop_edges != 1)] += init_edges.max() - 1
-            loop_edges[loop_edges == 0] = new_leaves[i][0]
-            loop_edges[loop_edges == 1] = new_leaves[i][1]
-            merged_edges = np.concatenate((init_edges, loop_edges))
-        else:
-            loop_edges[(loop_edges != 0) & (loop_edges != 1)] += merged_edges.max() - 1
-            loop_edges[loop_edges == 0] = new_leaves[i][0]
-            loop_edges[loop_edges == 1] = new_leaves[i][1]
-            merged_edges = np.concatenate((merged_edges, loop_edges))
+        loop_edges[(loop_edges != 0) & (loop_edges != 1)] += merged_edges.max() - 1
+        ix0, ix1 = loop_edges == 0, loop_edges == 1
+        loop_edges[ix0],loop_edges[ix1] = new_leaves[i][0], new_leaves[i][1]
+        merged_edges = np.concatenate((merged_edges, loop_edges))
     merged_nodep = np.concatenate((init_nodes_pos, *new_nodep))
 
     ### optionally refit the entire graph ###
@@ -731,17 +707,9 @@ def findPaths(
             cycle_nodes=cycle_nodes,
         )
 
-        (
-            merged_nodep,
-            _,
-            _,
-            _,
-            _,
-            _,
-            _,
-        ) = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
-            X, merged_nodep, ElasticMatrix, PointWeights=weights, FixNodesAtPoints=[],
-        )
+        merged_nodep = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
+        X, merged_nodep, ElasticMatrix, PointWeights=weights, FixNodesAtPoints=[])[0]
+
         # check intersection
         if merged_nodep.shape[1] == 2:
             intersect = not (
@@ -763,6 +731,8 @@ def findPaths(
     _PG["addLoopsdict"] = dict(
         new_edges=new_edges,
         new_nodep=new_nodep,
+        merged_nodep=merged_nodep,
+        merged_edges=merged_edges,
         new_leaves=new_leaves,
         new_part=new_part,
         new_energy=new_energy,
@@ -814,9 +784,15 @@ def addPath(
     _PG = deepcopy(PG)
     if "projection" in _PG.keys():
         del _PG["projection"]
+
     init_nodes_pos, init_edges = _PG["NodePositions"], _PG["Edges"][0]
+    init_edges_max = init_edges.max()
 
     # --- Init parameters, variables
+    if not isinstance(target,int):
+        init_nodes_pos = np.vstack((init_nodes_pos,target))
+        init_edges_max+= 1
+        target = len(init_nodes_pos)-1
     if Mu is None:
         Mu = _PG["Mu"]
     if Lambda is None:
@@ -844,7 +820,7 @@ def addPath(
     nodep, edges = PG_path["NodePositions"], PG_path["Edges"][0]
 
     _edges = edges.copy()
-    _edges[(edges != 0) & (edges != 1)] += init_edges.max() - 1
+    _edges[(edges != 0) & (edges != 1)] += init_edges_max - 1
     _edges[edges == 0] = source
     _edges[edges == 1] = target
     _merged_edges = np.concatenate((init_edges, _edges))
@@ -1008,7 +984,7 @@ def refitGraph(
     cycle_Lambda=None,
 ):
 
-    init_nodes_pos, init_edges = (PG["NodePositions"], PG["Edges"][0])
+    init_nodes_pos, init_edges = PG["NodePositions"], PG["Edges"][0]
 
     # --- Init parameters, variables
     if Mu is None:
@@ -1020,18 +996,16 @@ def refitGraph(
     if cycle_Lambda is None:
         cycle_Lambda = Lambda
 
-    # ---Modify node pos order (first nodes are fixed)
+    # ---Modify node pos values and order (first nodes are fixed)
     for k, v in shift_nodes_pos.items():
         init_nodes_pos[k] = v
+
     fix_nodes = sorted(list(shift_nodes_pos.keys()), reverse=True)
-    fix_order = np.arange(len(init_nodes_pos))
-    fix_edges = init_edges.copy()
-    for i, ifix in enumerate(fix_nodes):
-        n1, n2 = fix_order == i, fix_order == ifix
-        e1, e2 = fix_edges == i, fix_edges == ifix
-        fix_order[n1], fix_order[n2] = ifix, i
-        fix_edges[e1], fix_edges[e2] = ifix, i
+    fix_order = np.array(fix_nodes+[i for i in range(len(init_nodes_pos)) if i not in fix_nodes])
     fix_nodes_pos = init_nodes_pos[fix_order]
+    fix_edges = init_edges.copy()
+    for i, ifix in enumerate(fix_order):
+        fix_edges[init_edges == ifix] = i
 
     SquaredX = np.sum(X ** 2, axis=1, keepdims=1)
     part, part_dist = elpigraph.src.core.PartitionData(
@@ -1068,10 +1042,5 @@ def refitGraph(
     )
 
     # ---Revert to initial node ordering
-    for i, ifix in enumerate(fix_nodes):
-        e1, e2 = fix_edges == i, fix_edges == ifix
-        fix_edges[e1], fix_edges[e2] = ifix, i
-    new_nodes_pos = new_nodes_pos[fix_order]
-
-    PG["NodePositions"] = new_nodes_pos
+    PG["NodePositions"] = new_nodes_pos[np.argsort(fix_order)]
 
